package model;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Random;

import org.ejml.simple.SimpleMatrix;

public class McElieceHammingModel {

	// #########
	// Variables
	// #########

	private Random rnd;
	private HashMap<String, String> syndError;
	private SimpleMatrix scramblerMatrix, generatorMatrix, permutationMatrix, parityCheckMatrix, linearCode,
			iScramblerMatrix, iPermutationMatrix;
	private SimpleMatrix publicKey, plaintext, error, ciphertext;
	private SimpleMatrix unpCiphertext, syn, decoded, decrypted;
	private String[][] syndTable;

	public McElieceHammingModel() {
		rnd = new Random();
		syndError = new HashMap<String, String>();
		syndTable = new String[8][16];
	}

	// ######################
	// Acciones de generacion
	// ######################

	public void generateScramblerMatrix() {
		scramblerMatrix = getRandomScramblerMatrix(4);
		iScramblerMatrix = setMod2(scramblerMatrix.invert());
	}

	public void generateLinearCodeMatrices() {
		linearCode = getRandomCode(4, 3);
		generatorMatrix = getGeneratorMatrixFromCode(linearCode);
		parityCheckMatrix = getParityCheckMatrixFromCode(linearCode);
		syndError = loadSyndromeTable();
	}

	public void generatePermutationMatrix() {
		permutationMatrix = getRandomPermutationMatrix(7);
		iPermutationMatrix = permutationMatrix.invert();
	}

	public void generatePlainTextMessage() {
		plaintext = getRandomPlainText();
	}

	public void generateVectorError() {
		error = getRandomErrorVector(1);
	}

	private ArrayList<String> generateBinaryStrings(int N) {
		ArrayList<String> lista = new ArrayList<String>();
		String alfabet = "01";

		int iterations = (int) Math.pow(alfabet.length(), N);

		int pos[] = new int[N];
		pos[N - 1] = -1;
		for (int i = 0; i < iterations; i++) {
			pos[N - 1]++;
			for (int j = N - 1; j >= 0; j--) {
				if (pos[j] == alfabet.length()) {
					pos[j] = 0;
					pos[j - 1]++;
				} else
					break;
			}
			String word = "";
			for (int j = 0; j < N; j++) {
				word += alfabet.charAt(pos[j]);
			}

			lista.add(word);

		}

		ArrayList<String> sortedByHammingWeight = new ArrayList<String>();

		lista.stream().filter(n -> hammingWeight(n) == 0).forEach(n -> sortedByHammingWeight.add(n));
		lista.stream().filter(n -> hammingWeight(n) == 1).forEach(n -> sortedByHammingWeight.add(n));
		lista.stream().filter(n -> hammingWeight(n) == 2).forEach(n -> sortedByHammingWeight.add(n));
		lista.stream().filter(n -> hammingWeight(n) == 3).forEach(n -> sortedByHammingWeight.add(n));
		lista.stream().filter(n -> hammingWeight(n) == 4).forEach(n -> sortedByHammingWeight.add(n));
		lista.stream().filter(n -> hammingWeight(n) == 5).forEach(n -> sortedByHammingWeight.add(n));
		lista.stream().filter(n -> hammingWeight(n) == 6).forEach(n -> sortedByHammingWeight.add(n));
		lista.stream().filter(n -> hammingWeight(n) == 7).forEach(n -> sortedByHammingWeight.add(n));

		return sortedByHammingWeight;
	}

	private ArrayList<SimpleMatrix> generateBinarySimpleMatrix(int N) {
		ArrayList<SimpleMatrix> simp = new ArrayList<SimpleMatrix>();
		generateBinaryStrings(N).stream().forEach(n -> simp.add(getMatrixFromVector(n)));
		return simp;
	}

	// #################################
	// Funciones con resultado aleatorio
	// #################################

	private SimpleMatrix getRandomCode(int row, int column) {
		SimpleMatrix s = new SimpleMatrix(row, column);

		String[] code = { "110", "011", "111", "101" };

		Collections.shuffle(Arrays.asList(code));

		for (int i = 0; i < row; i++) {
			s.set(i, 0, Double.parseDouble("" + code[i].charAt(0)));
			s.set(i, 1, Double.parseDouble("" + code[i].charAt(1)));
			s.set(i, 2, Double.parseDouble("" + code[i].charAt(2)));
		}

		return s;

	}

	private SimpleMatrix getRandomErrorVector(int t) {
		SimpleMatrix s = new SimpleMatrix(1, 7);

		for (int i = 0; i < t; i++) {
			int col = rnd.nextInt(7);
			s.set(0, col, 1);
		}

		return s;

	}

	private SimpleMatrix getRandomPlainText() {
		return getRandomBinaryMatrix(1, 4);
	}

	private SimpleMatrix getRandomScramblerMatrix(int n) {
		SimpleMatrix s = getRandomBinaryMatrix(n, n);

		// int attempt = 1;
		while (!isInvertible(s)) {
			// System.out.println("intento:" + attempt);

			s = getRandomBinaryMatrix(n, n);

			// attempt++;
		}

		return s;
	}

	private SimpleMatrix getRandomBinaryMatrix(int n, int m) {
		SimpleMatrix s = new SimpleMatrix(n, m);

		for (int i = 0; i < s.numRows(); i++) {
			for (int j = 0; j < s.numCols(); j++) {
				s.set(i, j, rnd.nextInt(2));
			}
		}

		return setMod2(s);
	}

	private SimpleMatrix getRandomPermutationMatrix(int n) {
		SimpleMatrix s = new SimpleMatrix(n, n);

		ArrayList<Integer> lista = new ArrayList<Integer>();

		for (int i = 0; i < n; i++) {
			lista.add(i);
		}

		Collections.shuffle(lista);

		for (int i = 0; i < s.numRows(); i++) {
			s.set(i, lista.get(i), 1);
		}

		return s;
	}

	// ############################
	// Funciones de calculo interno
	// ############################

	private SimpleMatrix getParityCheckMatrixFromCode(SimpleMatrix c) {
		return c.transpose().concatColumns(SimpleMatrix.identity(c.getMatrix().getNumCols()));
	}

	private SimpleMatrix getGeneratorMatrixFromCode(SimpleMatrix c) {
		return SimpleMatrix.identity(c.getMatrix().getNumRows()).concatColumns(c);
	}

	public void encrypt() {
		ciphertext = setMod2(plaintext.mult(publicKey).plus(error));
	}

	public void decrypt() {
		decrypted = setMod2(decoded.extractMatrix(0, 1, 0, 4).mult(iScramblerMatrix));
	}

	public void correctError() {
		String s = printMatrix(syn);

		if (!s.equals("000")) {
			SimpleMatrix fix = getMatrixFromVector(syndError.get(s));
			decoded = setMod2(unpCiphertext.plus(fix));
		} else {
			decoded = unpCiphertext;
		}

	}

	public void getSyndromeFromCiphertext() {
		syn = setMod2(unpCiphertext.mult(parityCheckMatrix.transpose()));
	}

	public void getCipherWithInversePermutation() {
		unpCiphertext = setMod2(ciphertext.mult(iPermutationMatrix));
	}

	public void getPublicKey(SimpleMatrix S, SimpleMatrix G, SimpleMatrix P) {
		publicKey = setMod2(S.mult(G).mult(P));
	}

	public SimpleMatrix setMod2(SimpleMatrix s) {
		for (int i = 0; i < s.numRows(); i++) {
			for (int j = 0; j < s.numCols(); j++) {
				if (s.get(i, j) == -1.0) {
					s.set(i, j, 1);
				} else {
					s.set(i, j, s.get(i, j) % 2);
				}
			}
		}
		return s;
	}

	private boolean isInvertible(SimpleMatrix s) {
		try {
			s = s.invert();

			for (int i = 0; i < s.numRows(); i++) {
				for (int j = 0; j < s.numCols(); j++) {
					if (s.get(i, j) % 1 != 0) {
						return false;
					}
				}
			}

			return true;
		} catch (Exception e) {
			return false;
		}

	}

	private HashMap<String, String> loadSyndromeTable() {
		ArrayList<String> lista = generateBinaryStrings(4);

		SimpleMatrix[][] standardArray = new SimpleMatrix[8][16];

		int col = 0;
		for (String s : lista) {
			SimpleMatrix s2 = getMatrixFromVector(s);
			standardArray[0][col] = setMod2(s2.mult(generatorMatrix));
			syndTable[0][col] = printMatrix(standardArray[0][col]);
			col++;
		}

		for (int i = 1; i < standardArray.length; i++) {
			SimpleMatrix m = getMinimum(standardArray);

			for (int j = 0; j < standardArray[i].length; j++) {
				standardArray[i][j] = setMod2(standardArray[0][j].plus(m));
				syndTable[i][j] = printMatrix(standardArray[i][j]);

			}

		}

		HashMap<String, String> syndromes = new HashMap<String, String>();

		for (int i = 1; i < standardArray.length; i++) {
			SimpleMatrix s3 = setMod2(standardArray[i][0].mult(parityCheckMatrix.transpose()));
			syndromes.put(printMatrix(s3), printMatrix(standardArray[i][0]));
		}

		return syndromes;

	}

	private int hammingWeight(String o2) {
		int unos = 0;
		for (char k : o2.toCharArray()) {
			if (k == '1') {
				unos++;
			}
		}
		return unos;
	}

	private SimpleMatrix getMinimum(SimpleMatrix[][] standardArray) {
		SimpleMatrix m = new SimpleMatrix(1, 7);

		ArrayList<SimpleMatrix> binaryVectors = generateBinarySimpleMatrix(7);

		for (SimpleMatrix vector : binaryVectors) {
			boolean exists = false;
			for (int i = 0; i < standardArray.length; i++) {

				for (SimpleMatrix codeword : standardArray[i]) {
					if (isEqual(vector, codeword)) {
						exists = true;
					}
				}

			}

			if (!exists) {
				m = vector;
				break;
			}

		}

		return m;
	}

	private static boolean isEqual(SimpleMatrix s2, SimpleMatrix m2) {
		try {

			for (int i = 0; i < s2.numRows(); i++) {
				for (int j = 0; j < s2.numCols(); j++) {
					if (s2.get(i, j) != m2.get(i, j)) {
						return false;
					}

				}
			}
		} catch (Exception e) {
			return false;
		}
		return true;
	}

	private SimpleMatrix getMatrixFromVector(String word) {

		double[][] mat = new double[1][word.length()];

		for (int i = 0; i < word.length(); i++) {
			if (word.charAt(i) == '0') {
				mat[0][i] = 0.0;
			} else {
				mat[0][i] = 1.0;
			}
		}

		return new SimpleMatrix(mat);
	}

	private static String printMatrix(SimpleMatrix mat) {
		StringBuilder buffer = new StringBuilder();
		int m = mat.numRows();
		int n = mat.numCols();
		for (int i = 0; i < m; i++) {
			for (int j = 0; j < n; j++) {
				buffer.append(numToString(mat.get(i, j)));
			}
			if (m > 1 && i < m - 1) {
				buffer.append("\n");
			}
		}

		return buffer.toString();
	}

	private static String numToString(double arg) {
		String temp = Double.toString(arg).replace('E', 'e');
		if (temp.endsWith(".0"))
			temp = temp.substring(0, temp.length() - 2);
		return temp;
	}

	public void printMatlabMatrix(SimpleMatrix s) {
		StringBuilder sb = new StringBuilder();
		sb.append("[ ");
		sb.append("\n");
		for (int i = 0; i < s.numRows(); i++) {
			for (int j = 0; j < s.numCols(); j++) {
				int ss = (int) s.get(i, j);
				sb.append(ss + " ");
			}
			sb.append(";");
			sb.append("\n");

		}
		sb.append("]");

		System.out.println(sb.toString());
	}

	public String printMatrixToGUI(SimpleMatrix s) {
		StringBuilder sb = new StringBuilder();

		for (int i = 0; i < s.numRows(); i++) {
			for (int j = 0; j < s.numCols(); j++) {
				int ss = (int) s.get(i, j);
				sb.append(ss + " ");
			}
			sb.append("\n");
		}

		return sb.toString();
	}

	// #################
	// Getters y setters
	// #################

	public SimpleMatrix getScramblerMatrix() {
		return scramblerMatrix;
	}

	public void setScramblerMatrix(SimpleMatrix s) {
		scramblerMatrix = s;
	}

	public String[][] getSyndTable() {
		return syndTable;
	}

	public void setSyndTable(String[][] syndTable) {
		this.syndTable = syndTable;
	}

	public SimpleMatrix getGeneratorMatrix() {
		return generatorMatrix;
	}

	public void setGeneratorMatrix(SimpleMatrix g) {
		generatorMatrix = g;
	}

	public SimpleMatrix getPermutationMatrix() {
		return permutationMatrix;
	}

	public void setPermutationMatrix(SimpleMatrix p) {
		permutationMatrix = p;
	}

	public SimpleMatrix getParityCheckMatrix() {
		return parityCheckMatrix;
	}

	public void setParityCheckMatrix(SimpleMatrix h) {
		parityCheckMatrix = h;
	}

	public SimpleMatrix getLinearCode() {
		return linearCode;
	}

	public void setLinearCode(SimpleMatrix c) {
		linearCode = c;
	}

	public SimpleMatrix getCiphertext() {
		return ciphertext;
	}

	public void setCiphertext(SimpleMatrix y) {
		this.ciphertext = y;
	}

	public SimpleMatrix getPlaintext() {
		return plaintext;
	}

	public void setPlaintext(SimpleMatrix x) {
		this.plaintext = x;
	}

	public SimpleMatrix getError() {
		return error;
	}

	public void setError(SimpleMatrix e) {
		this.error = e;
	}

	public SimpleMatrix getPublicKeyMatrix() {
		return publicKey;
	}

	public void setPublicKeyMatrix(SimpleMatrix pk) {
		publicKey = pk;
	}

	public SimpleMatrix getUnpermutedCiphertext() {
		return unpCiphertext;
	}

	public void setUnpermutedCiphertext(SimpleMatrix yp) {
		this.unpCiphertext = yp;
	}

	public SimpleMatrix getSyndrome() {
		return syn;
	}

	public void setSyndrome(SimpleMatrix syn) {
		this.syn = syn;
	}

	public SimpleMatrix getDecodedText() {
		return decoded;
	}

	public void setDecodedText(SimpleMatrix ypp) {
		this.decoded = ypp;
	}

	public SimpleMatrix getDecryptedMessage() {
		return decrypted;
	}

	public void setDecryptedMessage(SimpleMatrix xx) {
		this.decrypted = xx;
	}

	public SimpleMatrix getInverseScramblerMatrix() {
		return iScramblerMatrix;
	}

	public void setInverseScramblerMatrix(SimpleMatrix sI) {
		iScramblerMatrix = sI;
	}

	public SimpleMatrix getInversePermutationMatrix() {
		return iPermutationMatrix;
	}

	public void setInversePermutationMatrix(SimpleMatrix pI) {
		iPermutationMatrix = pI;
	}

	// ###############
	// Main de pruebas
	// ###############

	public static void main(String[] args) {

		for (int i = 0; i < 100; i++) {

			McElieceHammingModel f = new McElieceHammingModel();

			// Genero las matrices S, G y P

			f.generateScramblerMatrix();
			f.generateLinearCodeMatrices();
			f.generatePermutationMatrix();

			// Genero la clave publica G' = SGP
			//
			f.getPublicKey(f.getScramblerMatrix(), f.getGeneratorMatrix(), f.getPermutationMatrix());

			// Genero un mensaje x

			f.generatePlainTextMessage();

			// Genero un vector de error e de tamano t

			f.generateVectorError();

			// Genero un mensaje cifrado y = xG'+e
			f.encrypt();

			// Genero un "despermutado" del mensaje cifrado y' = yP-1

			f.getCipherWithInversePermutation();

			// Calculo el sindrome de y' y lo busco en la tabla de sindromes
			// para corregir el error

			f.getSyndromeFromCiphertext();
			f.correctError();

			// Genero el mensaje descifrado x = y''(4)S-1

			f.decrypt();

			if (isEqual(f.getPlaintext(), f.getDecryptedMessage())) {
				System.out.println("Bien");
			} else {
				System.out.println("Mal");

				System.out.println(printMatrix(f.getScramblerMatrix()));
				System.err.println(f.getInverseScramblerMatrix());
				System.out.println();
				System.out.println(printMatrix(f.getPermutationMatrix()));
				System.out.println();
				System.out.println(printMatrix(f.getPlaintext()));
				System.out.println(printMatrix(f.getError()));
				System.out.println(printMatrix(f.getCiphertext()));

				System.out.println(printMatrix(f.getUnpermutedCiphertext()));

				System.out.println(printMatrix(f.getSyndrome()));
				System.out.println(printMatrix(f.getDecodedText()));

				System.out.println(printMatrix(f.getDecryptedMessage()));
			}

		}
	}

}
