package model;

import org.ejml.simple.SimpleMatrix;

import de.flexiprovider.api.Registry;
import de.flexiprovider.api.SecureRandom;
import de.flexiprovider.common.math.codingtheory.GF2mField;
import de.flexiprovider.common.math.codingtheory.GoppaCode;
import de.flexiprovider.common.math.codingtheory.GoppaCode.MaMaPe;
import de.flexiprovider.common.math.codingtheory.PolynomialGF2mSmallM;
import de.flexiprovider.common.math.codingtheory.PolynomialRingGF2m;
import de.flexiprovider.common.math.linearalgebra.GF2Matrix;
import de.flexiprovider.common.math.linearalgebra.GF2Vector;
import de.flexiprovider.common.math.linearalgebra.Permutation;
import de.flexiprovider.pqc.ecc.ECCKeyGenParameterSpec;

public class NiederreiterGoppaModel {

	// #########
	// Variables
	// #########

	private SimpleMatrix scramblerMatrix, generatorMatrix, permutationMatrix, parityCheckMatrix, linearCode,
			iScramblerMatrix, iPermutationMatrix;
	private SimpleMatrix publicKey, plaintext, error, ciphertext;
	private SimpleMatrix unsCiphertext, syn, decoded, decrypted;
	private String[][] syndTable;

	private SecureRandom random;
	private int n, m = 4, t = 2;
	private int fieldPoly;
	private GF2mField field;
	private PolynomialGF2mSmallM gp;
	private PolynomialRingGF2m ring;
	private PolynomialGF2mSmallM[] sqRootMatrix;
	private GF2Matrix hPrime, h, SI;
	private MaMaPe mmp;
	private Permutation P, PI;
	private GF2Vector pe;
	private GF2Vector x, c, e;
	private GF2Vector syndrome;

	int k = 8;

	public NiederreiterGoppaModel() {
		this.random = (random != null) ? random : Registry.getSecureRandom();

		ECCKeyGenParameterSpec mParams = new ECCKeyGenParameterSpec(m, t);

		m = mParams.getM();
		n = mParams.getN();
		t = mParams.getT();

		fieldPoly = mParams.getFieldPoly();
		field = new GF2mField(m, fieldPoly);

		gp = new PolynomialGF2mSmallM(field, t, PolynomialGF2mSmallM.RANDOM_IRREDUCIBLE_POLYNOMIAL, random);
		ring = new PolynomialRingGF2m(field, gp);
		sqRootMatrix = ring.getSquareRootMatrix();

		syndTable = new String[8][16];
	}

	// ######################
	// Acciones de generacion
	// ######################

	public void generateScramblerMatrix() {
		SI = mmp.getFirstMatrix();
		scramblerMatrix = gf2matrixToSimpleMatrix((GF2Matrix) SI.computeInverse());
		iScramblerMatrix = gf2matrixToSimpleMatrix(SI);
	}

	public void generateLinearCodeMatrices() {

		h = GoppaCode.createCanonicalCheckMatrix(field, gp);
		mmp = GoppaCode.computeSystematicForm(h, random);
		hPrime = mmp.getSecondMatrix();

		parityCheckMatrix = gf2matrixToSimpleMatrix(h);
		

	}

	public void generatePermutationMatrix() {
		P = mmp.getPermutation();
		PI = P.computeInverse();

		permutationMatrix = getMatrixFromPermutation(P);
		iPermutationMatrix = getMatrixFromPermutation(PI);
	}

	public void generatePlainTextMessage() {
		x = new GF2Vector(n);
		x.setBit(random.nextInt(n));

		plaintext = gf2VectorToSimpleMatrix(x);
	}

	public void generateVectorError() {
		error = getRandomErrorVector(1);
	}

	// #################################
	// Funciones con resultado aleatorio
	// #################################

	private SimpleMatrix getRandomErrorVector(int t) {
		SimpleMatrix s = new SimpleMatrix(1, 7);

		for (int i = 0; i < t; i++) {
			int col = random.nextInt(7);
			s.set(0, col, 1);
		}

		return s;

	}

	// ############################
	// Funciones de calculo interno
	// ############################

	public void encrypt() {	
		c = (GF2Vector) hPrime.rightMultiplyRightCompactForm(x);
		ciphertext = gf2VectorToSimpleMatrix(c);
	}

	public void decrypt() {
		e = (GF2Vector) pe.multiply(P);
		decrypted = gf2VectorToSimpleMatrix(e);
	}

	public void correctError() {

		pe = pattersonDecoding(syndrome, field, gp, sqRootMatrix);

		decoded = gf2VectorToSimpleMatrix(pe);
	}

	public void getSyndromeFromCiphertext() {
		syn = setMod2(unsCiphertext.mult(parityCheckMatrix.transpose()));
	}

	public void getCipherWithInverseScramble() {
		syndrome = (GF2Vector) SI.rightMultiply(c);
		unsCiphertext = gf2VectorToSimpleMatrix(syndrome);
	}

	public void getPublicKey(SimpleMatrix S, SimpleMatrix H, SimpleMatrix P) {
		publicKey = gf2matrixToSimpleMatrix(hPrime.extendRightCompactForm());
	}

	public SimpleMatrix setMod2(SimpleMatrix s) {
		for (int i = 0; i < s.numRows(); i++) {
			for (int j = 0; j < s.numCols(); j++) {
				if (s.get(i, j) == -1.0) {
					s.set(i, j, 1);
				} else {
					s.set(i, j, s.get(i, j) % 2);
				}
			}
		}
		return s;
	}

	private static boolean isEqual(SimpleMatrix s2, SimpleMatrix m2) {
		try {

			for (int i = 0; i < s2.numRows(); i++) {
				for (int j = 0; j < s2.numCols(); j++) {
					if (s2.get(i, j) != m2.get(i, j)) {
						return false;
					}

				}
			}
		} catch (Exception e) {
			return false;
		}
		return true;
	}

	private SimpleMatrix getMatrixFromVector(String word) {

		double[][] mat = new double[1][word.length()];

		for (int i = 0; i < word.length(); i++) {
			if (word.charAt(i) == '0') {
				mat[0][i] = 0.0;
			} else {
				mat[0][i] = 1.0;
			}
		}

		return new SimpleMatrix(mat);
	}

	public void printMatlabMatrix(SimpleMatrix s) {
		StringBuilder sb = new StringBuilder();
		sb.append("[ ");
		sb.append("\n");
		for (int i = 0; i < s.numRows(); i++) {
			for (int j = 0; j < s.numCols(); j++) {
				int ss = (int) s.get(i, j);
				sb.append(ss + " ");
			}
			sb.append(";");
			sb.append("\n");

		}
		sb.append("]");

		System.out.println(sb.toString());
	}

	public String printMatrixToGUI(SimpleMatrix s) {
		StringBuilder sb = new StringBuilder();

		for (int i = 0; i < s.numRows(); i++) {
			for (int j = 0; j < s.numCols(); j++) {
				int ss = (int) s.get(i, j);
				sb.append(ss + " ");
			}
			sb.append("\n");
		}

		return sb.toString();
	}

	public SimpleMatrix gf2matrixToSimpleMatrix(GF2Matrix gf) {
		SimpleMatrix s = new SimpleMatrix(0, gf.getNumColumns());

		String[] st = gf.toString().split("\n");

		for (String h : st) {
			StringBuilder vector = new StringBuilder(h.replace(" ", "").split(":")[1]);

			s = s.concatRows(getMatrixFromVector(vector.toString()));

		}

		return s;
	}

	public SimpleMatrix gf2VectorToSimpleMatrix(GF2Vector gf) {
		return getMatrixFromVector(gf.toString());
	}

	public SimpleMatrix getMatrixFromPermutation(Permutation p) {
		SimpleMatrix s = new SimpleMatrix(p.getVector().length, p.getVector().length);

		int count = 0;
		for (int i : p.getVector()) {
			s.set(i, count, 1);
			count++;
		}

		return s;
	}

	public GF2Vector pattersonDecoding(GF2Vector syndVec, GF2mField field, PolynomialGF2mSmallM gp,
			PolynomialGF2mSmallM[] sqRootMatrix) {
		int n = 1 << field.getDegree();

		// the error vector
		GF2Vector errors = new GF2Vector(n);

		// if the syndrome vector is zero, the error vector is also zero
		if (!syndVec.isZero()) {
			// convert syndrome vector to polynomial over GF(2^m)
			PolynomialGF2mSmallM syndrome = new PolynomialGF2mSmallM(syndVec.toExtensionFieldVector(field));

			// compute T = syndrome^-1 mod gp
			PolynomialGF2mSmallM t = syndrome.modInverse(gp);

			// compute tau = sqRoot(T + X) mod gp
			PolynomialGF2mSmallM tau = t.addMonomial(1);
			tau = tau.modSquareRootMatrix(sqRootMatrix);

			// compute polynomials a and b satisfying a + b*tau = 0 mod gp
			PolynomialGF2mSmallM[] ab = tau.modPolynomialToFracton(gp);

			// compute the polynomial a^2 + X*b^2
			PolynomialGF2mSmallM a2 = ab[0].multiply(ab[0]);
			PolynomialGF2mSmallM b2 = ab[1].multiply(ab[1]);
			PolynomialGF2mSmallM xb2 = b2.multWithMonomial(1);
			PolynomialGF2mSmallM a2plusXb2 = a2.add(xb2);

			// normalize a^2 + X*b^2 to obtain the error locator polynomial
			int headCoeff = a2plusXb2.getHeadCoefficient();
			int invHeadCoeff = field.inverse(headCoeff);
			PolynomialGF2mSmallM elp = a2plusXb2.multWithElement(invHeadCoeff);

			// for all elements i of GF(2^m)
			for (int i = 0; i < n; i++) {
				// evaluate the error locator polynomial at i
				int z = elp.evaluateAt(i);
				// if polynomial evaluates to zero
				if (z == 0) {
					// set the i-th coefficient of the error vector
					errors.setBit(i);
				}
			}
		}

		return errors;
	}

	// #################
	// Getters y setters
	// #################

	public SimpleMatrix getScramblerMatrix() {
		return scramblerMatrix;
	}

	public void setScramblerMatrix(SimpleMatrix s) {
		scramblerMatrix = s;
	}

	public SimpleMatrix getGeneratorMatrix() {
		return generatorMatrix;
	}

	public void setGeneratorMatrix(SimpleMatrix g) {
		generatorMatrix = g;
	}

	public SimpleMatrix getPermutationMatrix() {
		return permutationMatrix;
	}

	public void setPermutationMatrix(SimpleMatrix p) {
		permutationMatrix = p;
	}

	public SimpleMatrix getParityCheckMatrix() {
		return parityCheckMatrix;
	}

	public void setParityCheckMatrix(SimpleMatrix h) {
		parityCheckMatrix = h;
	}

	public SimpleMatrix getLinearCode() {
		return linearCode;
	}

	public void setLinearCode(SimpleMatrix c) {
		linearCode = c;
	}

	public String[][] getSyndTable() {
		return syndTable;
	}

	public void setSyndTable(String[][] syndTable) {
		this.syndTable = syndTable;
	}

	public SimpleMatrix getCiphertext() {
		return ciphertext;
	}

	public void setCiphertext(SimpleMatrix y) {
		this.ciphertext = y;
	}

	public SimpleMatrix getPlaintext() {
		return plaintext;
	}

	public void setPlaintext(SimpleMatrix x) {
		this.plaintext = x;
	}

	public SimpleMatrix getError() {
		return error;
	}

	public void setError(SimpleMatrix e) {
		this.error = e;
	}

	public SimpleMatrix getPublicKeyMatrix() {
		return publicKey;
	}

	public void setPublicKeyMatrix(SimpleMatrix pk) {
		publicKey = pk;
	}

	public SimpleMatrix getUnscrambledCiphertext() {
		return unsCiphertext;
	}

	public void setUnscrambledCiphertext(SimpleMatrix yp) {
		this.unsCiphertext = yp;
	}

	public SimpleMatrix getSyndrome() {
		return syn;
	}

	public void setSyndrome(SimpleMatrix syn) {
		this.syn = syn;
	}

	public SimpleMatrix getDecodedText() {
		return decoded;
	}

	public void setDecodedText(SimpleMatrix ypp) {
		this.decoded = ypp;
	}

	public SimpleMatrix getDecryptedMessage() {
		return decrypted;
	}

	public void setDecryptedMessage(SimpleMatrix xx) {
		this.decrypted = xx;
	}

	public SimpleMatrix getInverseScramblerMatrix() {
		return iScramblerMatrix;
	}

	public void setInverseScramblerMatrix(SimpleMatrix sI) {
		iScramblerMatrix = sI;
	}

	public SimpleMatrix getInversePermutationMatrix() {
		return iPermutationMatrix;
	}

	public void setInversePermutationMatrix(SimpleMatrix pI) {
		iPermutationMatrix = pI;
	}

	// ###############
	// Main de pruebas
	// ###############

	public static void main(String[] args) {
		NiederreiterGoppaModel f = new NiederreiterGoppaModel();

		// Genero las matrices S, H y P

		f.generateLinearCodeMatrices();
		f.generateScramblerMatrix();
		f.generatePermutationMatrix();

		// Genero la clave publica H' = SHP
		// f.getPublicKey(f.getScramblerMatrix(), f.getParityCheckMatrix(),
		// f.getPermutationMatrix());

		for (int i = 0; i < 10; i++) {

			// Genero un mensaje x

			f.generatePlainTextMessage();

			// Genero un mensaje cifrado y = Hx'
			f.encrypt();

			// Calculo SI*c

			f.getCipherWithInverseScramble();

			// Recupero el mensaje de la tabla de sindromes

			f.correctError();

			// Genero el mensaje descifrado m = (P'*(Pm'))'
			f.decrypt();

			if (isEqual(f.getPlaintext(), f.getDecryptedMessage())) {
				System.out.println("Bien");
			} else {
				System.out.println("Mal");
			}

		}
	}

}
